# Determine the national qualifiers for NCAA XC.
#
# This software is free.


# How it works:
# For the rules, see the rulebook on the ncaa website.
# http://www.ncaa.org/sites/default/files/2017DIMWCC_PreChamps_DI_CrossCountry_20170920.pdf
#
# There is a "data" object that holds all the information describing where we
# currently are in the process.
# Each round, we get a list of eligible teams and pick the top team.
# If that would create a "push" scenario, we:
#   -- copy the current data object (everything that's been done so far)
#   -- modify the copy to tell it NOT to use the push on this team
#   -- recursively call our procedure to figure out the winner
#   -- If that team would have made the qualifying list, then we're done.
#   -- Otherwise, throw out that recursive call and select this team and the team
#      ahead of them with a push.
#
# The data object also contains a list of "messages" describing what's going on;
# these all get printed out at the end.


import os,sys,operator
import util

# --------------------------------

class Data:
  def __init__(self, teamsin_, curr_inds_, pushes_used_, points_, messages_, dont_push_, meets, regionals):
    self.teamsin = list(teamsin_)    # list of teams currently in
    self.curr_inds = dict(curr_inds_)    # current index in each regional we're considering (like 3rd or 4th)
    self.pushes_used = dict(pushes_used_)  # list of regions that we used the push in (so can't use it again)
    self.points = dict(points_)      # map team name to points scored
    self.messages = list(messages_)  # list of what to print out at the end
    self.dont_push = list(dont_push_)      # list of teams not to use the push on (because we're testing it out)
    self.meets = meets
    self.regionals = regionals
  def createCopy(self):
    return Data(self.teamsin, self.curr_inds, self.pushes_used, self.points, self.messages, self.dont_push, self.meets, self.regionals)


def removeB(teamname):
  if teamname.endswith(" B team") or teamname.endswith(" B Team"):
    return teamname[:-len(" B team")]
  return teamname


# pick a team from the given region at the given index
# if awardpoints if false, then don't award any points for beating that team
# return a copy of data updated with this team selected.
def selectTeam(data,region,ind,awardpoints=True):
  d2 = data.createCopy()
  selected = d2.regionals[region].teams[ind]
  d2.teamsin.append(selected)
  d2.curr_inds[region] += 1
  if not awardpoints:
    d2.messages.append(str(len(d2.teamsin)) + ": " + selected + " (#" + str(d2.curr_inds[region]) + " " + region + ") is selected, gives no points (pushed in)\n")
    return d2
  d2.messages.append(str(len(d2.teamsin)) + ": " + selected + " (#" + str(d2.curr_inds[region]) + " " + region + ") is selected")
  beatme = {}  # map meet to list of teams that beat me there
  for m,mobj in sorted(d2.meets.items()):
    date = mobj.date
    l = mobj.teams
    if not selected in l:
      continue
    beatmetemp = []
    # iterate through teams that beat me here, but don't bother listing a team if it's already
    #   been selected or if it's in the top 2 in its region
    for teamname in l[:l.index(selected)]:
      t = removeB(teamname)
      addme = True
      if t in d2.teamsin:
        addme = False
      # this kicks in when adding auto teams early
      for regname,regobj in d2.regionals.items():
        teamlist = regobj.teams
        if t == teamlist[0] or t == teamlist[1]:
          addme = False
      if addme:
        beatmetemp.append(t)
    if len(beatmetemp) > 0:
      beatme[m] = beatmetemp
  if len(beatme.keys()) == 0:
    d2.messages.append(", has no losses to teams remaining")
  else:
    d2.messages.append(", has losses to ")
    for m,l in beatme.items():
      d2.messages.append("(" + d2.meets[m].name + ") " + ", ".join(l) + "; ")
      for ot in l: # award the points!!
        d2.points[ot] += 1
  d2.messages.append("\n")
  return d2


# get list of teams currently eligible to be selected
# for each region, check whether we've already used a push
# and for each potentially pushing team, see if they're on the "no-push" list
def getEligible(data):
  ans = []
  for rname,robj in data.regionals.items():
    l = robj.teams
    if data.curr_inds[rname] < len(l):
      ans.append(l[data.curr_inds[rname]])
      if len(data.teamsin) < 30 and not data.pushes_used[rname] and data.curr_inds[rname]+1 < len(l):
        if l[data.curr_inds[rname]+1] not in data.dont_push:
          ans.append(l[data.curr_inds[rname]+1])
  return sorted(ans, key=lambda k: data.points[k], reverse=True)

# print out the current status of everything
def printPoints(el,data):
  data.messages.append("Teams under consideration by points:\n")
  pts_to_teams = {}
  for e in el:
    pts = data.points[e]
    if pts in pts_to_teams:
      pts_to_teams[pts].append(e)
    else:
      pts_to_teams[pts] = [e]
  for pts,teamlist in sorted(pts_to_teams.items(), reverse=True):
    data.messages.append("  " + str(pts) + " "*(3-len(str(pts))) + ", ".join(teamlist) + "\n")
  data.messages.append("  Others with points: ")
  all_by_pts = sorted(data.points.keys(), key=lambda k: data.points[k], reverse=True)
  others = [t for t in all_by_pts if t not in el and t not in data.teamsin and data.points[t] >= 1]
  other_strlist = [t + " (" + str(data.points[t]) + ")" for t in others]
  data.messages.append(", ".join(other_strlist))
  data.messages.append("\n")


# given a list of tied teams and a data object, attempt to resolve ties
# by eliminating teams that definitely should not be selected
# the first tiebreaker is head-to-head
#  cannot select X ahead of Y if X has a losing record vs Y
# second is common opponents, which I interpret as
#  cannot select X ahead of Y if, of the teams that both faced,
#  X has more wins
def resolveTies(oldtied,d):
  assert len(oldtied) > 0
  if len(oldtied) == 1:
    return oldtied[0]
  tied = list(oldtied)

  # HEAD TO HEAD
  d.messages.append("  Breaking ties between " + ", ".join(tied) + " using head-to-head\n")
  lost = []
  # compare every pair of teams
  for i in range(len(tied)):
    ti = tied[i]
    for j in range(i+1,len(tied)):
      tj = tied[j]
      score = 0
      relevant = []
      for m,mobj in list(d.meets.items()) + list(d.regionals.items()):
        date = mobj.date
        teams = mobj.teams
        if ti in teams and tj in teams:
          dscore = 1 if teams.index(ti) < teams.index(tj) else -1
          relevant.append((date, m, dscore))
          score += dscore
      relevant.sort()  # sorts by date, earliest to latest
      if len(relevant) == 0:
        d.messages.append("  No head-to-head tiebreaker for " + ti + " and " + tj + "\n")
        continue
      elif score < 0:
        d.messages.append("  " + ti + " loses head-to-head vs " + tj + ", is not selected")
        lost.append(ti)
      elif score > 0:
        d.messages.append("  " + tj + " loses head-to-head vs " + ti + ", is not selected")
        lost.append(tj)
      else:  # tie, take winner of latest matchup
        # note regionals have default date of year 3000, so they are always the most recent if applicable
        dscore = relevant[-1][2]
        if dscore < 0:
          d.messages.append("  " + ti + " loses head-to-head vs " + tj + " (most recent matchup), is not selected")
          lost.append(ti)
        else:
          d.messages.append("  " + tj + " loses head-to-head vs " + ti + " (most recent matchup), is not selected")
          lost.append(tj)
      d.messages.append(" (" + ", ".join([r[1] for r in relevant]) + ")\n")
  for t in lost:
    if t in tied:
      tied.remove(t)
  if len(tied) == 1:
    return tied[0]  # this team won the head-to-head tiebreakers
  elif len(tied) == 0:
    d.messages.append("  Head-to-head tiebreaker eliminated all teams; ignoring.\n")
    tied = list(oldtied)     # got some contradiction with head-to-head tiebreakers ...
  else:
    oldtied = list(tied)     # eliminated some of the teams; save the surviving set


  # COMMON OPPONENTS
  d.messages.append("  Breaking ties between " + ", ".join(tied) + " using common opponents\n")
  lost = []
  for i in range(len(tied)):
    ti = tied[i]
    for j in range(i+1,len(tied)):
      tj = tied[j]
      opps = {}  # maps teamname --> [i's wins, j's wins, times_faced_i, times_faced_j]
      for m,mobj in list(d.meets.items()) + list(d.regionals.items()):
        teams = mobj.teams
        if ti in teams:
          # add every team ahead of ti as having beaten ti
          for k in range(teams.index(ti)):
            if teams[k] not in opps:
              opps[teams[k]] = [0,0,0,0]
            opps[teams[k]][2] += 1  # they played i
          # add every team behind ti as having lost to ti
          for k in range(teams.index(ti)+1,len(teams)):
            if teams[k] not in opps:
              opps[teams[k]] = [0,0,0,0]
            opps[teams[k]][0] += 1  # i beat them
            opps[teams[k]][2] += 1  # they played i
        # do the same thing for tj
        if tj in teams:
          for k in range(teams.index(tj)):
            if teams[k] not in opps:
              opps[teams[k]] = [0,0,0,0]
            opps[teams[k]][3] += 1
          for k in range(teams.index(tj)+1,len(teams)):
            if teams[k] not in opps:
              opps[teams[k]] = [0,0,0,0]
            opps[teams[k]][1] += 1
            opps[teams[k]][3] += 1
      common_opps = {}
      for o,res in opps.items():
        if res[2] >= 1 and res[3] >= 1:
          common_opps[o] = res

      # now we have the list of common opponents
      if len(common_opps) == 0:
        d.messages.append("  No common opponents found for " + ti + ", " + tj)
      else:
        wins_i = sum([res[0] for o,res in common_opps.items()])
        wins_j = sum([res[1] for o,res in common_opps.items()])
        total_i = sum([res[2] for o,res in common_opps.items()])
        total_j = sum([res[3] for o,res in common_opps.items()])
        win_percent_i = float(wins_i) / float(total_i)
        win_percent_j = float(wins_j) / float(total_j)
        if win_percent_i > win_percent_j:
          d.messages.append("  " + tj + " loses common opponents vs " + ti + ", is not selected (win percentages: " + str(win_percent_i) + ", " + str(win_percent_j) + ")")
          lost.append(tj)
        elif win_percent_i < win_percent_j:
          d.messages.append("  " + ti + " loses common opponents vs " + tj + ", is not selected (win percentages: " + str(win_percent_j) + ", " + str(win_percent_i) + ")")
          lost.append(ti)
        else:
          d.messages.append("  Common opponents are tied for " + ti + ", " + tj + " (win percentages: " + str(win_percent_i) + ", " + str(win_percent_j) + ")")
        d.messages.append(" (common opponents: " + ', '.join([o for o,res in common_opps.items()]) + ")\n")

  for t in lost:
    if t in tied:
      tied.remove(t)
  if len(tied) == 1:
    return tied[0]  # got a winner!
  elif len(tied) == 0:
    d.messages.append("  Common opponent tiebreaker eliminated all teams; ignoring.\n")
    tied = list(oldtied)  # got a contradiction
  else:
    oldtied = list(tied)  # eliminated some teams; save the surviving set


  # HIGHEST PLACE AT REGIONALS
  d.messages.append("  Breaking ties between " + str(tied) + " using highest place at regionals.\n")
  places = [50]*len(tied)  # just picked 50 as a high default number
  for i,team in enumerate(tied):
    for rname,robj in d.regionals.items():
      reg = robj.teams
      if team in reg:
        places[i] = reg.index(team)
        break
  best_place = min(places)
  newtied = []
  for i,team in enumerate(tied):
    if places[i] == best_place:
      newtied.append(team)

  if len(newtied) == 1:
    d.messages.append("  Broke tie by selecting " + newtied[0] + " (highest regional finish).\n")
    return newtied[0]
  elif len(newtied) == 0:
    pass  # ?? shouldn't happen
  else:
    tied = list(newtied)


  # CLOSEST TO 2ND SCORE AT REGIONALS
  d.messages.append("  Breaking ties between " + str(tied) + " using closest score to 2nd-place in region.\n")
  scored_diffs = [0]*len(tied)
  best_diff = 10000
  for i,team in enumerate(tied):
    for rname,robj in d.regionals.items():
      reg = robj.teams
      if team in reg:
        diff = robj.team_scores[reg.index(team)] - robj.team_scores[1]
        scored_diffs[i] = diff
        best_diff = min(best_diff, diff)
  newtied = [t for i,t in enumerate(tied) if scored_diffs[i] == best_diff]
  if len(newtied) == 1:
    d.messages.append("  Broke tie by selecting " + newtied[0] + " (closest score to regional 2nd-place).\n")
    return newtied[0]
  elif len(newtied) == 0:
    pass  # ?? shouldn't happen
  else:
    tied = newtied

  d.messages.append("  Could not break ties among " + ", ".join(tied) + "!\n")
  return tied


# given a data object, a team ("winner") to be selected, and list of other eligible teams
# try to select this team
# if there is no push, then just select them and continue
# otherwise, hairy stuff
def doWinner(d,winner,el):
  regwin = [r for r in d.regionals.keys() if winner in d.regionals[r].teams][0]  # get the regional this team is in
  indwin = d.regionals[regwin].teams.index(winner)
  # if there is no push, select and continue
  if indwin == d.curr_inds[regwin]:
    d = selectTeam(d,regwin,indwin)
    return doSelection(d)
  # if not, then this is a push situation
  # first, we create a copy of the data object and continue running without this team being eligible
  # to do this, add it to the no-push list so it doesn't try to push in again later
  # we add the message that it gets in later on its own just because we assume
  # that it works (if not, we'll throw out this selection anyway)
  # then we call pickFrom(), which runs the entire rest of the process and returns
  # a data object that has all 31 teams selected.
  # we check if our team made it. If so, fine, return that data object; otherwise,
  # throw it out, do the push instead and continue with the process.
  tryme = list(el)
  tryme.remove(winner)
  d2 = d.createCopy()
  d2.messages.append(winner + " not selected with push (gets in later on their own)\n")
  d2.dont_push.append(winner)
  final = pickFrom(d2,tryme)
  if final is None: return None
  if winner in final.teamsin:
    return final  # they did get in on their own
  # they didn't get in on their own, so use the push and select both teams
  # the pushed team gives no points
  d = selectTeam(d,regwin,indwin-1,awardpoints=False)
  d = selectTeam(d,regwin,indwin)
  d.pushes_used[regwin] = True
  return doSelection(d)


# given a data object and list of eligible teams,
# pick a team, make a selection, and continue on with the rest of the process
def pickFrom(d,el):
  tied = [i for i in el if d.points[i]==d.points[el[0]]]
  winner = resolveTies(tied,d)
  # if resolveTies failed, it returns a list; we ask the user to manually break ties
  if isinstance(winner,list):
    w = winner[0]
    d2 = d.createCopy()
    d2.messages.append("Manually broke tie by picking " + w + "\n\n")
    return doWinner(d2, w, el)
#    print(''.join(d.messages) + "\n\n")
#    results = []
#    print("Couldn't break the tie, please pick from ", winner)
#    while True:
#      try:  # for python2
#        w = raw_input()
#      except NameError:
#        w = input()
#      if w in winner:
#        break
#      print("Not in list, select again")
#    d2 = d.createCopy()
#    d2.messages.append("Manually broke tie by picking " + w + "\n\n")
#    return doWinner(d2,w,el)
  d.messages.append("\n")
  return doWinner(d,winner,el)


# do the next step of the selection!!
def doSelection(d):
  if len(d.teamsin) == 31:
    return d
  el = getEligible(d)
  printPoints(el,d)
  return pickFrom(d,el)






# --------------------------------


# Start from a blank slate
def init_data(meets, regionals):
  curr_inds = {}
  pushes_used = {}
  points = {}
  for r,l in sorted(regionals.items()):
    curr_inds[r] = 0
    pushes_used[r] = False
    for t in l.teams:
      points[t] = 0
  for m,l in meets.items():
    for t in l.teams:
      points[t] = 0
  return Data([],curr_inds,pushes_used,points,[],[],meets,regionals)

 

# year is "2018" or so on (a string)
# gender is "m" or "w"
# regs is a dictionary mapping regional name to list of teams
# return team_list, output_txt
def main(year, gender, regs):
  meets = util.get_meets(util.get_season_results(year, gender))
  regionals = util.get_regionals(regs)
  data = init_data(meets, regionals)

  # Select automatically the top two in each region
  data.messages.append("Automatically selected:\n")
  for r,l in sorted(regs.items()):
    data = selectTeam(data,r,0)
    data = selectTeam(data,r,1)
  data.messages.append("\n")
  
  # Select at-large teams
  results_data = doSelection(data)
  if results_data is None:
    return "There was an error!!"
  else:
    slist = ["\nTeams selected:\n"]
    for i,t in enumerate(results_data.teamsin):
      slist.append("%4s %s\n" % ((i+1),t))
    slist.append("\n")
    slist.append("\n")
    slist.extend(results_data.messages)
  return results_data.teamsin, ''.join(slist)


if __name__ == "__main__":
  # Takes two arguments, [m or w] and [year]
  if len(sys.argv) != 3 or sys.argv[1] not in ["m","w"]:
    print("Usage:\n    python3 quals.py m yyyy\n    or\n    python3 quals.py w yyyy\n")
    exit(0)
  gender = sys.argv[1]
  year = sys.argv[2]
  regs = util.get_regional_results(year, gender)
  teamlist, s = main(year, gender, regs)
  print(s)


